@@ -9,7 +9,8 @@ from urllib.parse import urljoin, urlparse
 import requests
 from bs4 import BeautifulSoup
 
-from autoscraper.utils import get_random_str, unique_hashable, unique_stack_list, ResultItem
+from autoscraper.utils import get_random_str, unique_hashable, unique_stack_list, \
+    ResultItem, FuzzyText, get_non_rec_text, text_match
 
 
 class AutoScraper(object):
@@ -118,10 +119,10 @@ class AutoScraper(object):
         return attrs
 
     @staticmethod
-    def _child_has_text(child, text, url):
+    def _child_has_text(child, text, url, text_fuzz_ratio):
         child_text = child.getText().strip()
 
-        if text == child_text:
+        if text_match(text, child_text, text_fuzz_ratio):
             parent_text = child.parent.getText().strip()
             if child_text == parent_text:
                 return False
@@ -129,12 +130,17 @@ class AutoScraper(object):
             child.wanted_attr = None
             return True
 
+        if text_match(text, get_non_rec_text(child), text_fuzz_ratio):
+            child.is_non_rec_text = True
+            child.wanted_attr = None
+            return True
+
         for key, value in child.attrs.items():
             if not isinstance(value, str):
                 continue
 
             value = value.strip()
-            if text == value:
+            if text_match(text, value, text_fuzz_ratio):
                 child.wanted_attr = key
                 return True
 
@@ -147,13 +153,14 @@ class AutoScraper(object):
 
         return False
 
-    def _get_children(self, soup, text, url):
+    def _get_children(self, soup, text, url, text_fuzz_ratio):
         text = text.strip()
         children = reversed(soup.findChildren())
-        children = [x for x in children if self._child_has_text(x, text, url)]
+        children = [x for x in children if self._child_has_text(x, text, url, text_fuzz_ratio)]
         return children
 
-    def build(self, url=None, wanted_list=None, wanted_dict=None, html=None, request_args=None, update=False):
+    def build(self, url=None, wanted_list=None, wanted_dict=None, html=None, request_args=None,
+              update=False, text_fuzz_ratio=1.0):
         """
         Automatically constructs a set of rules to scrape the specified target[s] from a web page.
             The rules are represented as stack_list.
@@ -184,9 +191,12 @@ class AutoScraper(object):
             If True, new learned rules will be added to the previous ones.
             If False, all previously learned rules will be removed.
 
+        text_fuzz_ratio: float in range [0, 1], optional, defaults to 1.0
+            The fuzziness ratio threshold for matching the wanted contents.
+
         Returns:
         --------
-        None
+        List of similar results
         """
 
         soup = self._get_soup(url=url, html=html, request_args=request_args)
@@ -206,7 +216,7 @@ class AutoScraper(object):
             wanted_list += wanted_items
 
             for wanted in wanted_items:
-                children = self._get_children(soup, wanted, url)
+                children = self._get_children(soup, wanted, url, text_fuzz_ratio)
 
                 for child in children:
                     result, stack = self._get_result_for_child(child, soup, url)
@@ -217,11 +227,8 @@ class AutoScraper(object):
         result_list = [item.text for item in result_list]
         result_list = unique_hashable(result_list)
 
-        if all(w in result_list for w in wanted_list):
-            self.stack_list = unique_stack_list(self.stack_list)
-            return result_list
-
-        return None
+        self.stack_list = unique_stack_list(self.stack_list)
+        return result_list
 
     @classmethod
     def _build_stack(cls, child, url):
@@ -248,7 +255,9 @@ class AutoScraper(object):
 
         wanted_attr = getattr(child, 'wanted_attr', None)
         is_full_url = getattr(child, 'is_full_url', False)
-        stack = dict(content=content, wanted_attr=wanted_attr, is_full_url=is_full_url)
+        is_non_rec_text = getattr(child, 'is_non_rec_text', False)
+        stack = dict(content=content, wanted_attr=wanted_attr, is_full_url=is_full_url,
+                     is_non_rec_text=is_non_rec_text)
         stack['url'] = url if is_full_url else ''
         stack['hash'] = hashlib.sha256(str(stack).encode('utf-8')).hexdigest()
         stack['stack_id'] = 'rule_' + get_random_str(4)
@@ -256,12 +265,14 @@ class AutoScraper(object):
 
     def _get_result_for_child(self, child, soup, url):
         stack = self._build_stack(child, url)
-        result = self._get_result_with_stack(stack, soup, url)
+        result = self._get_result_with_stack(stack, soup, url, 1.0)
         return result, stack
 
     @staticmethod
-    def _fetch_result_from_child(child, wanted_attr, is_full_url, url):
+    def _fetch_result_from_child(child, wanted_attr, is_full_url, url, is_non_rec_text):
         if wanted_attr is None:
+            if is_non_rec_text:
+                return get_non_rec_text(child)
             return child.getText().strip()
 
         if wanted_attr not in child.attrs:
@@ -272,14 +283,30 @@ class AutoScraper(object):
 
         return child.attrs[wanted_attr]
 
-    def _get_result_with_stack(self, stack, soup, url, **kwargs):
+    @staticmethod
+    def _get_fuzzy_attrs(attrs, attr_fuzz_ratio):
+        attrs = dict(attrs)
+        for key, val in attrs.items():
+            if isinstance(val, str) and val:
+                val = FuzzyText(val, attr_fuzz_ratio)
+            elif isinstance(val, (list, tuple)):
+                val = [FuzzyText(x, attr_fuzz_ratio) if x else x for x in val]
+            attrs[key] = val
+        return attrs
+
+    def _get_result_with_stack(self, stack, soup, url, attr_fuzz_ratio, **kwargs):
         parents = [soup]
         stack_content = stack['content']
         contain_sibling_leaves = kwargs.get('contain_sibling_leaves', False)
         for index, item in enumerate(stack_content):
             children = []
             for parent in parents:
-                found = parent.findAll(item[0], item[1], recursive=False)
+
+                attrs = item[1]
+                if attr_fuzz_ratio < 1.0:
+                    attrs = self._get_fuzzy_attrs(attrs, attr_fuzz_ratio)
+
+                found = parent.findAll(item[0], attrs, recursive=False)
                 if not found:
                     continue
 
@@ -293,29 +320,35 @@ class AutoScraper(object):
 
         wanted_attr = stack['wanted_attr']
         is_full_url = stack['is_full_url']
-        result = [ResultItem(self._fetch_result_from_child(i, wanted_attr, is_full_url, url),
+        is_non_rec_text = stack.get('is_non_rec_text', False)
+        result = [ResultItem(self._fetch_result_from_child(i, wanted_attr,
+                              is_full_url, url, is_non_rec_text),
                               getattr(i, 'child_index', 0)) for i in parents]
         result = [x for x in result if x.text]
         return result
 
-    def _get_result_with_stack_index_based(self, stack, soup, url, **kwargs):
+    def _get_result_with_stack_index_based(self, stack, soup, url, attr_fuzz_ratio, **kwargs):
         p = soup.findChildren(recursive=False)[0]
         stack_content = stack['content']
         for index, item in enumerate(stack_content[:-1]):
             content = stack_content[index + 1]
-            p = p.findAll(content[0], content[1], recursive=False)
+            attrs = content[1]
+            if attr_fuzz_ratio < 1.0:
+                attrs = self._get_fuzzy_attrs(attrs, attr_fuzz_ratio)
+            p = p.findAll(content[0], attrs, recursive=False)
             if not p:
                 return []
             idx = min(len(p) - 1, item[2])
             p = p[idx]
 
         result = [ResultItem(self._fetch_result_from_child(
-            p, stack['wanted_attr'], stack['is_full_url'], url), getattr(p, 'child_index', 0))]
+            p, stack['wanted_attr'], stack['is_full_url'], url, stack['is_non_rec_text']),
+            getattr(p, 'child_index', 0))]
         result = [x for x in result if x.text]
         return result
 
     def _get_result_by_func(self, func, url, html, soup, request_args, grouped,
-                            group_by_alias, unique, **kwargs):
+                            group_by_alias, unique, attr_fuzz_ratio, **kwargs):
         if not soup:
             soup = self._get_soup(url=url, html=html, request_args=request_args)
 
@@ -331,7 +364,7 @@ class AutoScraper(object):
             if not url:
                 url = stack.get('url', '')
 
-            result = func(stack, soup, url, **kwargs)
+            result = func(stack, soup, url, attr_fuzz_ratio, **kwargs)
 
             if not grouped and not group_by_alias:
                 result_list += result
@@ -366,8 +399,8 @@ class AutoScraper(object):
         return dict(grouped_result)
 
     def get_result_similar(self, url=None, html=None, soup=None, request_args=None,
-                           grouped=False, group_by_alias=False, unique=None, keep_order=False,
-                           contain_sibling_leaves=False):
+                           grouped=False, group_by_alias=False, unique=None, attr_fuzz_ratio=1.0,
+                           keep_order=False, contain_sibling_leaves=False):
         """
         Gets similar results based on the previously learned rules.
 
@@ -396,6 +429,9 @@ class AutoScraper(object):
                 False for grouped results.
             If set to True, will remove duplicates from returned result list.
 
+        attr_fuzz_ratio: float in range [0, 1], optional, defaults to 1.0
+            The fuzziness ratio threshold for matching html tag attributes.
+
         keep_order: bool, optional, defaults to False
             If set to True, the results will be ordered as they are present on the web page.
 
@@ -410,11 +446,12 @@ class AutoScraper(object):
 
         func = self._get_result_with_stack
         return self._get_result_by_func(func, url, html, soup, request_args, grouped,
-                                         group_by_alias, unique, keep_order=keep_order,
+                                         group_by_alias, unique, attr_fuzz_ratio,
+                                         keep_order=keep_order,
                                          contain_sibling_leaves=contain_sibling_leaves)
 
     def get_result_exact(self, url=None, html=None, soup=None, request_args=None,
-                         grouped=False, group_by_alias=False, unique=None):
+                         grouped=False, group_by_alias=False, unique=None, attr_fuzz_ratio=1.0):
         """
         Gets exact results based on the previously learned rules.
 
@@ -443,6 +480,9 @@ class AutoScraper(object):
                 False for grouped results.
             If set to True, will remove duplicates from returned result list.
 
+        attr_fuzz_ratio: float in range [0, 1], optional, defaults to 1.0
+            The fuzziness ratio threshold for matching html tag attributes.
+
         Returns:
         --------
         List of exact results scraped from the web page.
@@ -451,9 +491,10 @@ class AutoScraper(object):
 
         func = self._get_result_with_stack_index_based
         return self._get_result_by_func(func, url, html, soup, request_args, grouped,
-                                        group_by_alias, unique)
+                                        group_by_alias, unique, attr_fuzz_ratio)
 
-    def get_result(self, url=None, html=None, request_args=None, grouped=False, group_by_alias=False):
+    def get_result(self, url=None, html=None, request_args=None, grouped=False,
+                   group_by_alias=False, unique=None, attr_fuzz_ratio=1.0):
         """
         Gets similar and exact results based on the previously learned rules.
 
@@ -478,6 +519,13 @@ class AutoScraper(object):
             If set to True, the result will be a dictionary with the rule alias as keys
                 and a list of scraped data per alias as values.
 
+        unique: bool, optional, defaults to True for non grouped results and
+                False for grouped results.
+            If set to True, will remove duplicates from returned result list.
+
+        attr_fuzz_ratio: float in range [0, 1], optional, defaults to 1.0
+            The fuzziness ratio threshold for matching html tag attributes.
+
         Returns:
         --------
         Pair of (similar, exact) results.
@@ -485,8 +533,10 @@ class AutoScraper(object):
         """
 
         soup = self._get_soup(url=url, html=html, request_args=request_args)
-        similar = self.get_result_similar(url=url, soup=soup, grouped=grouped, group_by_alias=group_by_alias)
-        exact = self.get_result_exact(url=url, soup=soup, grouped=grouped, group_by_alias=group_by_alias)
+        args = dict(url=url, soup=soup, grouped=grouped, group_by_alias=group_by_alias,
+                                        unique=unique, attr_fuzz_ratio=attr_fuzz_ratio)
+        similar = self.get_result_similar(**args)
+        exact = self.get_result_exact(**args)
         return similar, exact
 
     def remove_rules(self, rules):
